import time

import numpy as np
from matplotlib import pyplot as plt

import iupitermag as im

plt.style.use("dark_background")


def bench(label: str, fn, n: int, repeat: int = 5, warmup: int = 1) -> float:
    # warmup
    for _ in range(warmup):
        fn()

    times = []
    for _ in range(repeat):
        t0 = time.perf_counter()
        fn()
        times.append(time.perf_counter() - t0)

    med = float(np.median(times))
    return med


def bench_for_num(N):
    rng = np.random.default_rng(0)
    positions = np.empty((N, 3), dtype=np.float64)
    positions[:, 0] = rng.uniform(5.0, 35.0, size=N)  # r
    positions[:, 1] = rng.uniform(0.0, np.pi, size=N)  # theta
    positions[:, 2] = rng.uniform(0.0, 2.0 * np.pi, size=N)  # phi

    field = im.InternalField("JRM33")

    def case_calc_field_loop() -> np.ndarray:
        out = np.empty((N, 3), dtype=np.float64)
        for i in range(N):
            r, theta, phi = positions[i]
            out[i, :] = field.calc_field(float(r), float(theta), float(phi))
        return out

    def case_map_calc_field() -> np.ndarray:
        return np.asarray(field.map_calc_field(positions))

    def case_parmap_calc_field() -> np.ndarray:
        return np.asarray(field.parmap_calc_field(positions))

    print("Benchmark: InternalField('JRM33')")
    print(f"N={N:,}")
    print()

    med1 = bench("1) calc_field (loop)", case_calc_field_loop, N)
    med2 = bench("2) map_calc_field", case_map_calc_field, N)
    med3 = bench("3) parmap_calc_field", case_parmap_calc_field, N)
    return (med1, med2, med3)


if __name__ == "__main__":
    len_num = 1000
    num_points = np.geomspace(1, 100_00, len_num, dtype=int)

    medians1 = np.empty(len_num)
    medians2 = np.empty(len_num)
    medians3 = np.empty(len_num)

    for i, n in enumerate(num_points):
        medians1[i], medians2[i], medians3[i] = bench_for_num(n)

    fig, ax = plt.subplots(1, 1, figsize=(6, 3), dpi=200)

    ax.loglog(num_points, medians1, lw=1.0, label="Python loop")
    ax.loglog(num_points, medians2, lw=1.0, label="map_calc_field")
    ax.loglog(num_points, medians3, lw=1.0, label="parmap_calc_field")
    ax.legend(loc="upper left", frameon=False)
    ax.set_xlabel("num points")
    ax.set_ylabel("time [s]")
    plt.show()
    fig.savefig("images/benchmark.png", facecolor="k", bbox_inches="tight")
    plt.show()
    plt.close(fig)



import numpy as np

import iupitermag


def test_internal_field():

    x = 10.0
    y = 0.0
    z = 0.0

    pos = np.array([np.sqrt(x**2 + y**2 + z**2), np.arccos(z / x), np.arctan2(y, x)])

    internal_field = iupitermag.InternalField("JRM09", degree=10)

    b_calc = internal_field.calc_field(pos[0], pos[1], pos[2])
    b_expected = np.array([-131.3754, 400.6375, -24.2054])
    assert np.allclose(b_expected, b_calc, rtol=1e-3)


def test_currentsheet_field():
    x = 20.2356
    y = 1.31
    z = -6.51

    currentsheet_field = iupitermag.CurrentSheetField("CON2020")

    b_calc = currentsheet_field.calc_field_xyz(x, y, z)
    b_expected = np.array([-31.20, 1.53, 23.60])
    assert np.allclose(b_expected, b_calc, rtol=1e-3)



import iupitermag._core as _iu

from .field import Field


class CurrentSheetField(Field):
    def __init__(self, typefield, params={}, integration_type="analytic"):
        self._field = _iu.PyCurrentSheetField(typefield, params, integration_type)



import numpy as np
from matplotlib import pyplot as plt

import iupitermag as im

plt.style.use("dark_background")


def calc_field_iupitermag(pos):
    internalField = im.InternalField("JRM33")
    return internalField.map_calc_field(pos)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    f = 1 / 15.4
    a = 1.0
    b = a
    c = (1 - f) * a

    theta_arr = np.linspace(0, np.pi, 100)
    phi_arr = np.linspace(0, 2 * np.pi, 100)

    theta, phi = np.meshgrid(theta_arr, phi_arr, indexing="ij")

    theta = np.ravel(theta)
    phi = np.ravel(phi)

    pos_xyz = np.zeros((100 * 100, 3))
    pos_xyz[:, 0] = a * np.sin(theta) * np.cos(phi)
    pos_xyz[:, 1] = b * np.sin(theta) * np.sin(phi)
    pos_xyz[:, 2] = c * np.cos(theta)

    pos = np.zeros((100 * 100, 3))
    pos[:, 0] = np.sqrt(np.sum(pos_xyz**2, axis=1))
    pos[:, 1] = np.arccos(pos_xyz[:, 2] / pos[:, 0])
    pos[:, 2] = np.arctan2(pos_xyz[:, 1], pos_xyz[:, 0])

    mag_array = calc_field_iupitermag(pos)

    Bmag = np.sqrt(np.sum(mag_array**2, axis=1)).reshape((100, 100))

    fig = plt.figure(dpi=200, figsize=(6, 3))
    ax = plt.subplot(111, projection="mollweide")

    longitude = phi_arr - np.pi
    latitude = np.pi / 2 - theta_arr

    mp = ax.contourf(longitude, latitude, Bmag, cmap="inferno")
    ax.set_title(r"$|\mathbf{B}|$")
    # ax.set_xlabel(r"Lon [$^\circ$]")
    ax.set_ylabel(r"Lat [$^\circ$]")
    ax.grid(True, color="0.5", lw=0.25)
    ax.set_xticklabels([])

    cbar = fig.colorbar(mp)
    cbar.set_label("nT")
    fig.savefig("images/jupiter_surfacefield.png", facecolor="k", bbox_inches="tight")

    plt.show()
    plt.close(fig)



import numpy as np

import iupitermag


def test_trace():

    start_pos = np.array([[-10.0, 2.0, 3.0]])

    internal_field = iupitermag.InternalField("JRM33", degree=10)
    currentsheet_field = iupitermag.CurrentSheetField("CON2020")

    traces = iupitermag.trace_field_to_planet(
        start_pos, internal_field, currentsheet_field
    )

    # The point traced to the planet in the northen hemisphere.
    first_expected = np.array([-0.52819934, -0.01459706, 0.77033209])
    assert np.allclose(first_expected, traces[0][0, :], rtol=1e-3)



import iupitermag._core as _iu


class Field:
    _field: _iu.PyInternalField | _iu.PyCurrentSheetField

    def __init__(self):
        """
        Base class for all field types. Contains generic routines like
        calc_field, map_calc_field, trace_field, etc.
        """
        pass

    def calc_field(self, r, theta, phi):
        """
        Calculate the field strength at a location defined in a spherical
        planetocentric coordinate system.

        Args:
            r (float): Radius in planetary radii
            theta (float): Co-latitude in radians
            phi (float): Azimuth in radians

        Returns:
            field (np.array): Magnetic field components [Br, Btheta, Bphi]
        """
        return self._field.calc_field(r, theta, phi)

    def calc_field_xyz(self, x, y, z):
        """
        Calculate the field strength at a location defined by (x, y, z) in a Cartesian
        coordinate system.
        """
        return self._field.calc_field_xyz(x, y, z)

    def get_params(self):
        """
        Get the parameters of the field. Varies for each field type.
        """
        return self._field.get_params()

    def map_calc_field(self, positions):
        return self._field.map_calc_field(positions)

    def parmap_calc_field(self, positions):
        return self._field.parmap_calc_field(positions)

    def map_calc_field_xyz(self, positions):
        return self._field.map_calc_field_xyz(positions)

    def parmap_calc_field_xyz(self, positions):
        return self._field.parmap_calc_field_xyz(positions)
